library(diversitree)
library(hisse)

correct_ultrametricized_trees <- function(tree){
  depths_of_nodes <- node.depth.edgelength(tree)
  number_of_tips <- length(tree$tip.label)
  depths_of_tips <- depths_of_nodes[1:number_of_tips]
  max_depth <- max(depths_of_tips)
  edges <- tree$edge
  for (i in 1:length(tree$edge.length)){
    if (tree$edge[i, 2] <= number_of_tips){
      tip <- tree$edge[i, 2]
      tree$edge.length[i] <- tree$edge.length[i] + (max_depth - depths_of_tips[tip])
    }
  }
  return (tree)
}

##################################################################################################
get_tree_name <- function(method, phyml, sum_of_logs){
  name <- "dated_treepl.newick"
  if (method == "NPRS"){
    name <- "dated_tree_NPRS.newick"
  }else if (method == "PATHd8"){
    name <- "dated_tree_pathd8.newick"
    
  }else if (method == "r8s"){
    name <- "r8s_dated_tree.newick"
  }else if (method == 'mrbayes'){
    if (sum_of_logs){
      name <- "MAP_tree.tre"
      
    }else{
      name <- "parsemb_map_tree.tre"
      
    }
    
  }
  if(phyml == 1){
    name <- paste(c("phyML_", name), collapse = "")
  }
  return (name)
}
##################################################################################################
fit_bisse_models <- function(output_dir, phyml, states_path, method, sum_of_logs){
  tree_file_path_name <- get_tree_name(method, phyml, sum_of_logs)
  false_positive <- 0
  file_extention <- ""
  tree_path <- NULL
  tree <- NULL
  states <- readRDS(states_path)
  dir_lrt <- paste(c(output_dir, get_name_of_HiSSE_dir(method, sum_of_logs)), collapse = "/")
  if(!dir.exists(dir_lrt)){
    dir.create(dir_lrt)
  }

  
  if (phyml == 1){
    file_extention <- "phyML_"
  }
  file_extention <- paste(c(file_extention, method, '_'), collapse = "")
  tree_path <-  paste(c(output_dir, tree_file_path_name), collapse = "/")
  print(tree_path)
  tree <- read.tree(tree_path)

  tree <- correct_ultrametricized_trees(tree)
  species <- names(states)
  states <- unname(states)
  data_for_hisse <- data.frame(species, states)
  row.names(data_for_hisse) <- species
  
  
  trans_matrix_hisse_CID <- TransMatMaker(hidden.states=TRUE)
  trans_matrix_hisse_CID <- ParDrop(trans_matrix_hisse_CID, c(3,5,8,10))
  # defining three different rates for trans. matrix:
  to.change <- cbind(c(1,3), c(2,4))#0->1
  trans_matrix_hisse_CID[to.change] = 1
  to.change <- cbind(c(2,4), c(1,3))#1->0
  trans_matrix_hisse_CID[to.change] = 2
  to.change <- cbind(c(1,3,2,4), c(3,1,4,2))#A->B and B->A
  trans_matrix_hisse_CID[to.change] = 3
  
  
  full_hisse <- hisse(tree, data_for_hisse, f=c(1,1), hidden.states=TRUE, turnover.anc=c(1,2,3,4),
                           eps.anc=c(1,2,3,4), trans.rate=trans_matrix_hisse_CID, output.type="raw")
  CID_2 <- hisse(tree, data_for_hisse, f=c(1,1), hidden.states=TRUE, turnover.anc=c(1,1,2,2),
                 eps.anc=c(1,1,2,2), trans.rate=trans_matrix_hisse_CID, output.type="raw")
  trans_matrix_for_bisse <- TransMatMaker(hidden.states=FALSE)
  
  bisse_SDD <- hisse(tree, data_for_hisse, f=c(1,1), hidden.states=FALSE, turnover.anc=c(1,2,0,0),
                     eps.anc=c(1,2,0,0), trans.rate=trans_matrix_for_bisse, output.type="raw")
  bisse_null <- hisse(tree, data_for_hisse, f=c(1,1), hidden.states=FALSE, turnover.anc=c(1,1,0,0),
                      eps.anc=c(1,1,0,0), trans.rate=trans_matrix_for_bisse, output.type="raw")
  
  models_parameters_list <- mget(c("bisse_null", "bisse_SDD", "CID_2", "full_hisse"))
  
  ## specifying data frame
  models <- c("BiSSE_null", "BiSSE_SDD", "CID_2", "full_HiSSE")
  AIC_values <- c(bisse_null$AIC, bisse_SDD$AIC, CID_2$AIC, full_hisse$AIC)
  matrix_AIC <- matrix(nrow = 1, ncol = length(models), dimnames = list(c("AIC"), models))
  data_frame_AIC <- data.frame(matrix_AIC)
  data_frame_AIC[1,] <- AIC_values
  
  best_AIC_null <- min(bisse_null$AIC, CID_2$AIC)
  best_AIC_null_model <- names(data_frame_AIC[which(data_frame_AIC == min(data_frame_AIC["BiSSE_null"], data_frame_AIC["CID_2"]))])
  best_AIC_SDD <- min(bisse_SDD$AIC, full_hisse$AIC)
  best_AIC_SDD_model <- names(data_frame_AIC[which(data_frame_AIC == min(data_frame_AIC["BiSSE_SDD"], data_frame_AIC["full_HiSSE"]))])
  
  if (best_AIC_SDD <= best_AIC_null - 2){
    false_positive <- 1
  }
  

  binary_res_file <- paste(c(file_extention, "false_positives.rds"), collapse = "")
  data_frame_file <- paste(c(file_extention, "model_selection.csv"), collapse = "")
  best_model_file <- paste(c(file_extention, "best_model.rds"), collapse = "")

  false_pos_full_path <- paste(c(dir_lrt, binary_res_file), collapse = "/")
  model_selection_csv <- paste(c(dir_lrt, data_frame_file), collapse = "/")
  best_model_full_path <- paste(c(dir_lrt, best_model_file), collapse = "/")
  
  # save parameter estimates for each model
  for (i in 1:length(models)){
    model_name <- models[i]
    write_model_solution(dir_lrt, model_name, models_parameters_list[[i]], file_extention)
  }
  
  
  write.csv(data_frame_AIC, model_selection_csv)

  
  saveRDS(false_positive, false_pos_full_path)
  
}
############################################################################################################
write_model_solution <- function(parent_dir, model_name, model_solution, file_extention){
  model_solution_short_path_name <- paste(c(file_extention, model_name, "_solution.rds"), collapse = "")
  full_path <- paste(c(parent_dir, model_solution_short_path_name), collapse = "/")
  solution_contents <- model_solution$solution
  saveRDS(solution_contents, full_path)
  
}
############################################################################################################


get_name_of_HiSSE_dir <- function(method, sum_of_logs){
  if (method == "mrbayes"){
    if (sum_of_logs){
      return ('HiSSE_sum_ln')
    }else{
      return ('HiSSE_ln')
    }
    
  }
  return ('HiSSE')
}

############################################################################################################
sum_of_logs <- 0
args <- commandArgs(TRUE)
for( i in 1:length(args) ){
  eval(parse(text = args[[i]]))
}

fit_bisse_models(output_dir, phyml, states_path, method, sum_of_logs)

