library(phytools)
library(diversitree)
library(phangorn)

## resolve polytomies according to a given topology

resolve_all_polytomies_in_trees <- function(output_dir, number_of_replicates, phyML, list_of_r, list_of_taxa, start){
  extension <- ""
  if (phyML){
    extension <- "phyML_"
  }

  for (i in 1:length(list_of_r)){
    
    r_full_dir <-  paste(c(output_dir, list_of_r[i]), collapse ="/")
    for(j in 1:length(list_of_taxa)){

      size_full_dir <-  paste(c(r_full_dir, list_of_taxa[j]), collapse ="/")
      
      
      for (k in start:number_of_replicates){
        rep_full_dir <-  paste(c(size_full_dir, as.character(k), "r8s"), collapse ="/")
        dated_tree_with_polytomies <- paste(c(rep_full_dir, paste(c(extension, "r8s_dated_tree.newick"), collapse = "")), collapse = "/")
        path_for_prev_tree <-  paste(c(rep_full_dir, paste(c(extension, "r8s_dated_tree_poly.newick"), collapse = "")), collapse = "/")
        path_for_topology <- paste(c(size_full_dir, as.character(k), paste(c(extension, "tree.newick"), collapse = "")), collapse = "/")
        if(!file.exists(dated_tree_with_polytomies)){
          print(paste(c("doesn't exist", dated_tree_with_polytomies), collapse = " "))
          next
        }
        if (file.exists(path_for_prev_tree)){
          print(paste(c("run not finished ", path_for_prev_tree), collapse = ""))
          next


        }else{
          res <- resolve_polytomies(dated_tree_with_polytomies, path_for_topology, path_for_prev_tree)
          if (res == 0){
            return (0)
          }
          
        }
        
      }

    }
    
  }
  return(1)

}



######################################################################################################

resolve_polytomies <- function(tree_path, topology_path, path_to_rename){
  tree <- read.tree(tree_path)
  tree_topo <- read.tree(topology_path)
  if (is.binary.tree(tree) & is.rooted(tree)){
    print("tree is completely binary")
    write.tree(tree, path_to_rename)
    tree <- correct_ultrametricized_trees(tree)
    if (!is.null(tree$node.label)){
      tree$node.label <- NULL
    }
    write.tree(tree, tree_path)
    return (1)
    
  }
  resolved_tree <- resolve_tree(tree, tree_topo)
  
  if (KF.dist(resolved_tree, tree) > 0){
    print(paste(c("resolved tree has KF distance above 0 ", tree_path), collapse = ""))
    return(0)
  }else if(!is.binary.tree(resolved_tree) | !(is.rooted(resolved_tree))){
    print(paste(c("still has polytomies ", tree_path), collapse = ""))
    return (0)
  }else{
    write.tree(tree, path_to_rename)
    resolved_tree <- correct_ultrametricized_trees(resolved_tree)
    write.tree(resolved_tree, tree_path)
  }

  return (1)

}
###################################################################################################
correct_ultrametricized_trees <- function(tree){

  depths_of_nodes <- node.depth.edgelength(tree)
  number_of_tips <- length(tree$tip.label)
  depths_of_tips <- depths_of_nodes[1:number_of_tips]
  max_depth <- max(depths_of_tips)
  edges <- tree$edge
  for (i in 1:length(tree$edge.length)){
    if (tree$edge[i, 2] <= number_of_tips){
      tip <- tree$edge[i, 2]
      tree$edge.length[i] <- tree$edge.length[i] + (max_depth - depths_of_tips[tip])
    }
  }
  return (tree)
}
###################################################################################################
resolve_tree <- function(tree, tree_topo){
  branch_lengths <- rep(0, length(tree_topo$edge.length))
  nodes_tree <- get.descendants(1, tree)
  nodes_topo <- get.descendants(1, tree_topo)
  indices_branch_lengths_topo <- get_indices(nodes_topo, tree_topo)
  indices_branch_lengths_poly <- get_indices(nodes_tree, tree)
  descendants_list_topo <- get_descendants_list(nodes_topo, tree_topo)
  descendants_list_poly <- get_descendants_list(nodes_tree, tree)
  
  for (i in 2:length(nodes_tree)){
    for (j in 2:length(nodes_topo)){
      contained <- is.contained(descendants_list_poly[[i]], descendants_list_topo[[j]])
      if (contained == 2){
        branch_lengths[indices_branch_lengths_topo[j]] <- tree$edge.length[indices_branch_lengths_poly[i]]
      }
    }
  }
  tree$edge <- tree_topo$edge
  tree$edge.length <- branch_lengths
  tree$tip.label <- tree_topo$tip.label
  if (!is.null(tree$node.label)){
    tree$node.label <- NULL
  }
  tree$Nnode <- tree_topo$Nnode
  return (tree)
  
}
################################################################################################
get_descendants_list <- function (nodes, phy){
  lst_of_descendants <- list()
  for (i in 1:length(nodes)){
    if (nodes[i] <= length(phy$tip.label)){
      lst_of_descendants[[i]] <- c(phy$tip.label[nodes[i]])
    }
    #print(nodes[i])
    else{
      lst_of_descendants[[i]] <- phy$tip.label[get.descendants(nodes[i], phy, tips.only = TRUE)]
    }
    
  }
  return (lst_of_descendants)
}

#################################################################################################
get_indices <- function(nodes, phy){
  indices <- c(0)
  for (i in 1:length(nodes)){
    for (j in 1:length(phy$edge.length)){
      if (phy$edge[j,2] == nodes[i]){
        indices <- c(indices, j)
      }
      
    }
  }
  return (indices)
}


###################################################################################################
resolve_polytomy <- function(node_poly, node_topo, branch_lengths,  tree_topo, tree){


  if (node_poly <= length(tree$tip.label)){
    leaf <- tree$tip.label[node_poly]
    index_of_leaf <- match(c(leaf), tree_topo$tip.label)
    index_in_branch_lengths_poly <- get_index_in_branches(node_poly, tree$edge, length(tree$edge.length))
    print(index_of_leaf)
    return (c(branch_lengths, tree$edge.length[index_in_branch_lengths_poly], index_of_leaf))
  }else{
    additional_lst <- c()

    children_topo <- get_children(node_topo, tree_topo)
    children_poly <- get_children(node_poly, tree)
    children_poly_descendants <- getDescendantsList(children_poly, tree)
    
    for (i in 1:length(children_topo)){
      children_descendants <- tree_topo$tip.label[get.descendants(children_topo[i], tree_topo, tips.only = TRUE)]
      for (j in 1:length(children_poly_descendants)){
        current_poly_node_descendants <- children_poly_descendants[[j]]
        contained <- is.contained(current_poly_node_descendants, children_descendants)
        if (contained == 2){
          edge_to <- children_topo[i]
          index_in_branch_lengths_poly <- get_index_in_branches(children_poly[j], tree$edge, length(tree$edge.length))
          edge_length <- tree$edge.length[index_in_branch_lengths_poly]
          print(edge_to)
          
          additional_lst <- c(additional_lst, resolve_polytomy(children_poly[j], children_topo[i], branch_lengths, tree_topo, tree), edge_length, edge_to)
          
          
        }else if (contained == 1){
   
          
          grand_children <- get_children(children_topo[i], tree_topo)
          for (idx in 1:length(grand_children)){
            grand_children_descendants <- tree_topo$tip.label[get.descendants(grand_children[idx], tree_topo, tips.only = TRUE)]
            contained_in_grand <- is.contained(current_poly_node_descendants, grand_children_descendants)
            additional_lst <- c(additional_lst, resolve_polytomy(children_poly[j], grand_children[idx], branch_lengths, tree_topo, tree))
            
          }
          
          
        }
      }
    }
  }
  return (additional_lst)
}
###################################################################################################

getDescendantsList<- function (children, phy){
  lst_of_descendants <- list()
  for (i in 1:length(children)){
    if(is.null(children[i])){
      print("Oh!")
    }else if (children[i] <= length(phy$tip.label)){
      lst_of_descendants[[i]] <- phy$tip.label[children[i]]
    }
    lst_of_descendants[[i]] <- phy$tip.label[get.descendants(children[i], phy, tips.only = TRUE)]
  }
  return (lst_of_descendants)
  
}


###################################################################################################

nodes_in_post_order <- function(node, tree, nodes){
  if (node <= length(tree$tip.label)){
    return (c(nodes, node))
    
  }else{
    children <- get_children(node, tree)
    
    return(c(node, nodes_in_post_order(children[1], tree, nodes), nodes_in_post_order(children[2], tree, nodes)))
    
  }
}


####################################################################################################
get_index <- function (node, child, phy){
  for (i in 1:length(phy$edge.length)){
    if (phy$edge[i,1] == node & phy$edge[i, 2] == child){
      return (i)
    }
  }
  return (NULL)
}
####################################################################################################
get_index_in_branches <- function(index_of_leaf, edges_of_tree, num_of_branches){
  for (i in 1:num_of_branches){
    if (edges_of_tree[i, 2] == index_of_leaf){
      return (i)
    }
  }
  return (NULL)
  
}


###################################################################################################

get_children <- function (node, phy){

  children <- c()
  if (node <= length(phy$tip.label)){
    return (c(node))
  }
  for (i in 1:length(phy$edge.length)){
    if (phy$edge[i, 1] == node){
      children <- c(children, phy$edge[i,2])
    }
  }
  return (children)
}


is.contained <- function(des_poly, des_binary){
  result <- 0
  if (sum(is.element(des_poly, des_binary)) == length(des_poly)){
    result <- 1
    if (length(des_poly) == length(des_binary)){
      result <- 2
    }

  }
  return(result)
}

##############################################################################

args <- commandArgs(TRUE)
for( i in 1:length(args) ){
  eval(parse(text = args[[i]]))
}

resolve_all_polytomies_in_trees(output_dir, number_of_replicates, phyML, list_of_r, list_of_taxa, start)

