library(diversitree)
library(phytools)
source("rescale_depth.R")


###############################################################################################

find_proportions <- function(p_matrix, mapping, branch_length){
  num_states_per_branch <- nrow(mapping)
  if (is.null(num_states_per_branch)){
    print("ERROR is here")
  }
  
  #current_state = mapping[1,2]
  #start = 0
  
  for ( i in 1:num_states_per_branch){
    current_state <- mapping[i, 2]
    if (i != num_states_per_branch){
      
      p_matrix[current_state+1, 1] <-  p_matrix[current_state+1, 1] + mapping[i+1,1] - mapping[i,1]
      
      #current_state = mapping[i,2]
      #start = mapping[i,1]
    }else{
      p_matrix[current_state+1, 1] <-  p_matrix[current_state+1, 1] + branch_length - mapping[i,1]
    }
  }
  p_matrix[1,1] <- p_matrix[1,1]/ branch_length
  p_matrix[2,1] <- p_matrix[2,1]/ branch_length
  return (p_matrix)
}
#############################################################################################

# final rescaling according to tree length
rescale_according_to_length <- function(phy, sum_of_lengths, total_length){
  branch_lengths <- phy$edge.length
  for (i in 1:length(branch_lengths)){
    phy$edge.length[i] <- total_length*(phy$edge.length[i]/sum_of_lengths)
  }
  return(phy)
}


##########################################################################################
get_state_of_node <- function(phy, node, mapping){
  state_of_node<- mapping$node.state[node-length(phy$tip.label)]
  return(state_of_node)
  
}

##########################################################################################


get_tree_total_length <- function(phy){
  total <- 0
  branches <- phy$edge.length
  for ( i in 1:length(branches) ){
    total <- total + branches[i]
  }
  return(total)
}

###########################################################################################################
get_branches_to_alter <- function(phy, descendants){
  indicator_of_descendants <- rep(0, length(phy$edge.length))
  edges <- phy$edge
  for ( i in 1:length(phy$edge.length)){
    if ( edges[i,1] %in% descendants && edges[i, 2] %in% descendants){
      indicator_of_descendants[i] <- 1
    }
  }
  return (indicator_of_descendants)
  
}
###########################################################################################################

# rescale tree branch lengths according to r and B
change_branch_lengths <- function(phy, mapping, r, depth, node){
  
  descendants <- get.descendants(node, phy)
  relevant_mapping <- get_branches_to_alter(phy, descendants)
  
  tree <- scale_depth(phy, depth)## rescale the ultrametric tree to a specific depth ##
  tree_length <- get_tree_total_length(tree)## get length of tree ##
  
  #rescaled_tree <- change_branch_lengths(tree, mapping, r, tree_depth)
  
  edges <- phy$edge
  branch_lengths <- phy$edge.length  # because need to scale according to the original branch lengths
  sum_of_lengths = 0  # represents the tree length of the stretched tree
  
  for (i in 1:length(branch_lengths)){
    if (relevant_mapping[i] == 1){

      if(!is.null(mapping)){
        
        proportions_matrix <- matrix(c(0,0,0,0), nrow = 2, ncol = 2)
        
        proportions_matrix[2,2] = 1 # state 1, the [1,2] index of the matrix is state 0
        
        mapping_of_node <- mapping$history[[i]]
        
        proportions_matrix <- find_proportions(proportions_matrix, mapping_of_node, branch_lengths[i])
        
        tree$edge.length[i] <- tree$edge.length[i] * (proportions_matrix[2,1]*r + proportions_matrix[1,1])
        
      }else{
        tree$edge.length[i] <- ( tree$edge.length[i] * r )
      }
      

    }

    sum_of_lengths <- sum_of_lengths + tree$edge.length[i]
    
  }

  rescaled_tree <- rescale_according_to_length(tree, sum_of_lengths, tree_length)
  return(rescaled_tree)
  
}

####### Temporary functions #########################################


check_if_exists_zero_length_edge <- function(phy){
  branches <- phy$edge
  zero_length <- vector(mode="integer", length= nrow(branches))
  branch_lengths <- phy$edge.length
  number_of_branches <- nrow(branches)
  for (i in 1:number_of_branches){
    if (phy$edge.length[i] == 0){
      zero_length[i] = 1
    }
  }
  return (zero_length)
}

####################################################################

simulate_tree <- function(rates, taxa, root_state){
  tree <- NULL
  counter <- 0
  while(is.null(tree) & counter < 100){ # sometimes the resulted tree is null. This loop should fix it.
    tree <- tree.bisse(rates, max.t= Inf, max.taxa = taxa, x0=root_state)
    counter <- counter + 1
  }
  if (is.null(tree)){
    print("ERROR in creation of the tree!!!")
  }
  return (tree)
}
#########################################################################################################

get_children <- function(phy, node){
  children <- c(0,0)
  edges <- phy$edge
  number_of_edges <- length(phy$edge.length)
  counter <- 0
  for (i in 1:number_of_edges){
    if(counter == 2){
      break
    }
    if(phy$edge[i,1] == node){
      counter = counter + 1
      children[counter] = phy$edge[i,2]
    }
    
  }
  return (children)
}

#######################################################################################################
get_number_of_descendants <- function(phy){
  descendants <- rep(0, length(phy$node.label))
  for (i in 1:length(phy$node.label)){
    descendants[i] <- length(get.descendants(i + length(phy$tip.label), phy, tips.only = TRUE))
  }
  return (descendants)
}
#######################################################################################################
get_root_of_subtree <- function(phy, taxa, lower_bound, upper_bound){
  abs_lower_bound <- round(taxa * lower_bound)
  abs_upper_bound <- round(taxa * upper_bound)
  num_descendants_per_node <- get_number_of_descendants(phy)
  mean_range <- round((abs_upper_bound+abs_upper_bound)/2)
  min_dis <- NULL
  root_subtree <- NULL
  for (i in 1:length(phy$node.label)){
    if (num_descendants_per_node[i] > abs_upper_bound || num_descendants_per_node < abs_lower_bound){
      next
    }
    if(is.null(min_dis)){
      min_dis <- abs(num_descendants_per_node[i] - mean_range)
      root_subtree <- i
      
    }else{
      distance <- abs(num_descendants_per_node[i] - mean_range)
      if (distance < min_dis){
        min_dis <- distance
        root_subtree <- i
      }
    }
  }
  return(root_subtree)
}

########################################################################################################
main <- function(number_taxa, rates, r, tree_depth, dir_tree_rep, lower_bound, upper_bound, trait){

  if(as.integer(r) == 1){

    tree <- simulate_tree(rates, number_taxa, 0)

    write.tree(tree, paste(c(dir_tree_rep, "original_tree.newick"),collapse = "/"))
    
    if (trait == 1){
      mapping <- history.from.sim.discrete(tree, 0:1)
      saveRDS(mapping, paste(c(dir_tree_rep, "mapping.rds"),collapse = "/"))
      states <- tree$tip.state
    }

  }

  if(as.integer(r) > 1){
    print("r is higher than 1!\n")
    replicate_num <- basename(dir_tree_rep)
    taxa_num <- basename(dirname(dir_tree_rep))
    taxa_directory <- dirname(dir_tree_rep)
    r_directory <- dirname(taxa_directory)
    simulations_directory <- dirname(r_directory)
    tree <- read.tree(paste(c(simulations_directory,'1.0', taxa_num, replicate_num, "original_tree.newick"),collapse = "/"))
    if (trait == 1){
      mapping <- readRDS(paste(c(simulations_directory,'1.0', taxa_num, replicate_num, "mapping.rds"),collapse = "/"))
      states <- readRDS(paste(c(simulations_directory,'1.0', taxa_num, replicate_num, "states.rds"), collapse = "/"))
    }
  }
  root_of_subtree <- get_root_of_subtree(tree, number_taxa, lower_bound, upper_bound)
  
  if (trait == 1){
    rescaled_tree <- change_branch_lengths(tree, mapping, r, tree_depth, root_of_subtree)
    saveRDS(states, paste(c(dir_tree_rep, "states.rds"),collapse = "/"))
  }else{
    print("rescaling the tree....\n")
    rescaled_tree <- change_branch_lengths(tree, NULL, r, tree_depth, root_of_subtree)
    if (is.null(rescaled_tree)){
      print("the tree is null...\n")
    }else{
      print("some other reason...\n")
    }
  }

  
  write.tree(rescaled_tree, paste(c(dir_tree_rep, "tree.newick"),collapse = "/")) # rescaled true tree
  
  max_depth <- get_depth_of_tree(rescaled_tree)
  b_tree <- create_tree_with_outgroup(rescaled_tree, rates, max_depth)
  
  tree_for_phyML_usage_path <- paste(c(dir_tree_rep, "tree_for_phyML.newick"), collapse = "/")# contains outgroup (for phyML)
  write.tree(b_tree, tree_for_phyML_usage_path)
  
  
  
  
}

########################################################################################################

create_tree_with_outgroup <- function(rescaled_tree, rates, tree_depth){
  
  subtree_for_rooting <- simulate_tree(rates, 3, 0)
  
  # change species names in the dummy subtree
  tip_names <- c("*s1", "*s2", "*s3")
  descendants_option1 <- getDescendants(subtree_for_rooting, 4)
  descendants_option2 <- getDescendants(subtree_for_rooting, 5)
  tips_to_drop <- descendants_option2
  if (length(descendants_option1) > length(descendants_option2)){# non- root internal node is needed
    node_for_graft <- 5
  }else{
    node_for_graft <- 4
    tips_to_drop <- descendants_option1
  }
  
  subtree_for_rooting$tip.label <- tip_names
  attributes(subtree_for_rooting$tip.label)$names <- tip_names
  b_tree <- bind.tree(subtree_for_rooting, rescaled_tree, where = node_for_graft, interactive = FALSE)
  for(i in 1:length(tips_to_drop)){
    b_tree <- drop.tip(b_tree, tip_names[tips_to_drop[i]])  # leave only one outgroup species
  }
  root <- getMRCA(b_tree, b_tree$tip.label)
  indices <- get_edges_from_root(b_tree, root)
  b_tree$edge.length[indices[1]] <- tree_depth * 1.2 # branch to a leaf
  b_tree$edge.length[indices[2]] <- tree_depth * 0.2
  
  return (b_tree)
}
##############################################################################################################
get_edges_from_root <- function(phy, root){
  vector_of_indices <- c(0,0)
  for(i in 1:length(phy$edge.length)){
    if(phy$edge[i, 1] == root){
      if(phy$edge[i,2] <= length(phy$tip.label)){
        vector_of_indices[1] <- i # edge directed to a leaf
      }else{
        vector_of_indices[2] <- i # edge directed to internal node
      }
    }
  }
  return(vector_of_indices)
}


##############################################################################################################
invert_indices_to_names <- function(phy, list_of_indices){
  list_of_names <- rep("", length(list_of_indices))
  for (i in 1:length(list_of_indices)){
    list_of_names[i] <- find_name_of_node(phy, list_of_indices[i])
  }
  return(list_of_names)
}
##############################################################################################################
find_name_of_node <- function(phy, node_index){
  if (node_index <= length(phy$tip.label)){
    return (phy$tip.label[node_index])
  }
  return (phy$node.label[node_index-length(phy$tip.label)])
}

###############################################################################################################

change_tip_names <- function (subtree_size, number_of_tips_base_tree){
  tip_labels <- rep("", subtree_size)
  for (i in 1:subtree_size){
    tip_labels[i] <- paste(c("sp",as.character(i+number_of_tips_base_tree)), collapse = "")
  }
  return (tip_labels)
}

##############################################################################################################
get_branch_length <- function(phy, node){
  edges <- phy$edge
  index <- match(c(node), phy$edge[,2])
  branch_length <- phy$edge.length[index]
  return(branch_length)
}
#############################################################################################################
get_ancestor <- function(phy, node){
  edges <- phy$edge
  index <- match(c(node), phy$edge[,2])
  ancestor <- edges[index, 1]
  return(ancestor)
}





###############################################################################################################

args <- commandArgs(TRUE)
for( i in 1:length(args) ){
  eval(parse(text = args[[i]]))
  }
main(number_taxa, rates, r, tree_length, tree_rep_dir, lower_bound, upper_bound, trait)
